import functools
import inspect
import logging


def log(logger: logging.Logger = None):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func_logger = logger if logger is not None else logging.getLogger(func.__name__)

            try:
                args_names = inspect.signature(func).parameters.keys()
                args_dict = dict(zip(args_names, args))
                args_repr = [f"{k}={v}" for k, v in args_dict.items()]
                kwargs_repr = [f"{k}={v}" for k, v in kwargs.items()]
                signature = ", ".join(args_repr + kwargs_repr)
                func_logger.debug(f"function {func.__name__} called with args {signature}")
            except:
                func_logger.debug(f"function {func.__name__} called")

            try:
                result = func(*args, **kwargs)
                return result
            except Exception as e:
                func_logger.exception(f"Exception raised in {func.__name__}", exc_info=True)
                raise e

        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        wrapper.__module__ = func.__module__
        wrapper.__qualname__ = func.__qualname__

        return wrapper
    return decorator


# logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(lineno)d - %(message)s')
# @log()
# def my_func(qwer):
#     print("hello world")
# my_func("qqq")



from trest.rest_request import RESTRequest


def test_get_curl_string():
    request = RESTRequest(method='POST', url='https://api.ipify.org',
                          params={'param1': 'one', 'param2': 'two'},
                          headers={'header1': 'one', 'header2': 'four'},
                          body={'body1': 'one', 'body2': 'two'})

    expected = 'curl -X POST -H "header1: one" -H "header2: four" ' \
               '-d \'{"body1": "one", "body2": "two"}\' ' \
               'https://api.ipify.org?param1=one&param2=two'
    assert request.get_curl_string() == expected


# def test_get_curl_string_one_line():
#     request = RESTRequest(method='POST', url='https://api.ipify.org',
#                           params={'param1': 'one', 'param2': 'two'},
#                           headers={'header1': 'one', 'header2': 'four'},
#                           body={'body1': 'one', 'body2': 'two'})
#
#     expected = "curl --location --request POST https://api.ipify.org?param1=one&param2=two " \
#                "--header 'header1: one' --header 'header2: four' " \
#                "--data-raw '{\"body1\": \"one\", \"body2\": \"two\"}'"
#     assert request.get_curl_string(one_line=True) == expected






from trest.rest_response import RESTResponse


def test_get_string_repr():
    response = RESTResponse(url='https://api.ipify.org', method='GET', path='/foo',
                            request_headers={'header1': 'value1', 'header2': 'value2'},
                            request_body='{"body1":"value1"}', status_code=200,
                            reason='ok', response_headers={'header1': 'value1', 'header2': 'value2'},
                            response_body='{"body1":"value1"}', elapsed_time='0.123')

    expected = '200 ok 0.123\n\nHeaders:\nheader1: value1\nheader2: value2\n\n' \
               'Content:\n{\n    "body1": "value1"\n}'
    assert response.get_string_repr() == expected


def test_get_string_repr_one_line():
    response = RESTResponse(url='https://api.ipify.org', method='GET', path='/foo',
                            request_headers={'header1': 'value1', 'header2': 'value2'},
                            request_body='{"body1":"value1"}', status_code=200,
                            reason='ok', response_headers={'header1': 'value1', 'header2': 'value2'},
                            response_body='{"body1":"value1"}', elapsed_time='0.123')

    expected = '200 ok 0.123, Headers: header1: value1, header2: value2, ' \
               'Content: {"body1": "value1"}'
    assert response.get_string_repr(one_line=True) == expected



from mock_api_server import MockApiServer
from trest.rest_request import RESTRequest
from trest.configuration.config import Config


def test_print_multiline_request(capsys):
    Config.one_line_request = False
    Config.print_request_to_std_out = True

    request = RESTRequest(method='POST', url='https://www.example.com',
                          params={'key1': 'value1', 'key2': 'value2'},
                          headers={'Content-Type': 'application/json'},
                          body='{"key1": "value1", "key2": "value2"}')
    response = request.send()

    expected = 'curl -X POST \\\n' \
               '-H "Content-Type: application/json" \\\n' \
               '-d \'{\n    "key1": "value1",\n    "key2": "value2"\n}\' \\\n' \
               'https://www.example.com?key1=value1&key2=value2\n'
    actual = capsys.readouterr().out
    assert actual == expected


def test_print_oneline_request(capsys):
    Config.one_line_request = True
    Config.print_request_to_std_out = True

    request = RESTRequest(method='POST', url='https://www.example.com',
                          params={'key1': 'value1', 'key2': 'value2'},
                          headers={'Content-Type': 'application/json'},
                          body='{"key1": "value1", "key2": "value2"}')
    response = request.send()

    expected = 'curl -X POST -H "Content-Type: application/json" ' \
               '-d \'{"key1": "value1", "key2": "value2"}\' ' \
               'https://www.example.com?key1=value1&key2=value2\n'
    actual = capsys.readouterr().out
    assert actual == expected


def test_print_multiline_response(capsys):
    Config.one_line_response = False
    Config.print_response_to_std_out = True

    request = RESTRequest(method='POST', url='http://localhost:5000/api/return_request',
                          params={'key1': 'value1', 'key2': 'value2'},
                          headers={'Content-Type': 'application/json'},
                          body='{"key1": "value1", "key2": "value2"}')
    with MockApiServer():
        request.send()

    actual = capsys.readouterr().out
    assert len(actual) > 2
    assert actual.strip().count('\n') > 1


def test_print_oneline_response(capsys):
    Config.one_line_response = True
    Config.print_response_to_std_out = True

    request = RESTRequest(method='POST', url='http://localhost:5000/api/return_request',
                          params={'key1': 'value1', 'key2': 'value2'},
                          headers={'Content-Type': 'application/json'},
                          body='{"key1": "value1", "key2": "value2"}')
    with MockApiServer():
        request.send()

    actual = capsys.readouterr().out
    assert len(actual) > 2
    assert actual.strip().count('\n') <= 1



import inspect
import json
import logging
from typing import Dict

from trest.utils import is_json_string


class RESTResponse(object):
    _log = logging.getLogger(__name__)

    def __init__(self, url: str,
                 method: str,
                 path: str,
                 request_headers: Dict[str, str],
                 request_body: str,
                 status_code: int,
                 reason: str,
                 response_headers: Dict[str, str],
                 response_body: str,
                 elapsed_time: str):
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')

        self.url = url
        self.method = method
        self.path = path
        self.request_headers = request_headers
        self.request_body = request_body

        self.status_code = status_code
        self.reason = reason
        self.response_headers = response_headers
        self.response_body = response_body
        self.elapsed_time = elapsed_time

    def get_response_body_json(self):
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        if not is_json_string(self.response_body):
            raise TypeError(f'Response body {self.response_body} is not json')
        return json.loads(self.response_body)

    def get_request_json_repr(self):
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        json_response = {
            'status_code': self.status_code,
            'reason': self.reason,
            'headers': self.response_headers,
            'elapsed_time': self.elapsed_time,
            'body': self.response_body if not is_json_string(self.response_body) else json.loads(
                self.response_body)
        }
        json_request = {
            'url': self.url,
            'method': self.method,
            'path': self.path,
            'headers': self.request_headers,
            'body': self.request_body if not is_json_string(self.request_body) else json.loads(
                self.request_body),
            'response': json_response
        }
        return json_request

    def get_string_repr(self, one_line: bool = False) -> str:
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        result_string = f'{self.status_code} {self.reason} {self.elapsed_time}'

        if self.response_headers:
            if one_line:
                result_string = result_string + ', Headers: '
                headers = []
                for key, value in self.response_headers.items():
                    headers.append(f"{key}: {value}")
                result_string = result_string + ', '.join(headers)
            else:
                result_string = result_string + '\n\nHeaders:\n'
                headers = []
                for key, value in self.response_headers.items():
                    headers.append(f"{key}: {value}")
                result_string = result_string + '\n'.join(headers)

        if self.response_body:
            if one_line:
                result_string = result_string + ', Content: '
                body_string = self.response_body if not is_json_string(self.response_body) else json.dumps(
                    json.loads(self.response_body))
                result_string = result_string + body_string
            else:
                result_string = result_string + '\n\nContent:\n'
                body_string = self.response_body if not is_json_string(self.response_body) else json.dumps(
                    json.loads(self.response_body), indent=4)
                result_string = result_string + body_string

        return result_string



import inspect
import logging
from typing import Dict, List, Union

from trest.object_converter import ObjectConverter


class CurlConverter:
    _log = logging.getLogger(__name__)

    @classmethod
    def to_curl(cls, method: str,
                url: str,
                params: Dict[str, Union[str, List[str]]] = None,
                headers: Dict[str, str] = None,
                data: str = None,
                multiline: bool = False) -> str:
        """
        Generates Curl string representation
        :param method: http method
        :param url: url
        :param params: dictionary of url parameters
        :param headers: dict of headers
        :param data: data string
        :param multiline: if True, returns curl string in multiline format
        :return: curl string representation
        """
        cls._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                       f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        if multiline:
            curl_cmd = f'curl -X {method}'
        else:
            curl_cmd = f'curl -X {method}'

        if headers is not None:
            if multiline:
                headers = ' \\\n'.join(f'-H "{key}: {value}"' for key, value in headers.items())
                curl_cmd += f' \\\n{headers}'
            else:
                headers = ' '.join(f'-H "{key}: {value}"' for key, value in headers.items())
                curl_cmd += f' {headers}'

        if data is not None:
            if multiline:
                curl_cmd += f' \\\n-d \'{ObjectConverter.to_string(data, indent=4)}\''
            else:
                curl_cmd += f' -d \'{data}\''

        if params is not None:
            params_list = []
            for key, value in params.items():
                if type(value) == list:
                    for item in value:
                        params_list.append(f'{key}={item}')
                else:
                    params_list.append(f'{key}={value}')
            url = f'{url}?'+'&'.join(params_list)
        if multiline:
            curl_cmd += f' \\\n{url}'
        else:
            curl_cmd += f' {url}'
        return curl_cmd



from dataclasses import dataclass, field

import allure
from trest.rest_request import RESTRequest


@allure.title("Test get request")
def test_get_request():
    request = RESTRequest(method='GET', url='https://api.ipify.org')
    response = request.send()


@allure.title("Test post request with dataclass body")
def test_post_request_with_dataclass_body():
    @dataclass
    class Data:
        Id: int = field(default=None, metadata={'name': 'Id', 'required': False})
        Customer: str = field(default=None, metadata={'name': 'Customer', 'required': False})
        Quantity: int = field(default=None, metadata={'name': 'Quantity', 'required': False})
        Price: float = field(default=None, metadata={'name': 'Price', 'required': False})

    headers = {'Content-Type': 'application/json'}
    body = Data(1, 'aaa', 2, 3.33)

    request = RESTRequest(method='POST', url='https://reqbin.com/echo/post/json', headers=headers, body=body)
    response = request.send()



import inspect
import json
import logging
from dataclasses import is_dataclass
from typing import TypeVar, Generic, Dict, Union, List

import allure
import requests
from jto import JTOConverter
from requests import Response
from trest.curl_converter import CurlConverter
from trest.object_converter import ObjectConverter

from trest.rest_response import RESTResponse
from trest.configuration.config import Config
from trest.utils import is_json_string

T = TypeVar('T')


class RESTRequest(Generic[T]):
    _log = logging.getLogger(__name__)

    def __init__(self, method: str,
                 url: str,
                 params: Dict[str, Union[str, List[str]]] = None,
                 headers: Dict[str, str] = None,
                 body: T = None,
                 hooks: dict = None,
                 timeout: float = None):
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')

        self.method = method
        self.url = url
        self.params = params
        self.headers = headers
        self.body: T = body
        self.hooks = hooks
        self.timeout = timeout

        self.response: Union[RESTResponse, None] = None

    def get_json(self, indent: int = None) -> str:
        """
        Get json representation of the request.
        :param indent: integer of indentation
        :return:
        """
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        return json.dumps(self, default=vars, indent=indent)

    def _convert_body_to_string(self, indent: int = None) -> str:
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        body_string = ObjectConverter.to_string(self.body, indent=indent)
        return body_string

    def get_curl_string(self) -> str:
        """
        Get curl string representation of the request.
        :return:
        """
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        result_string = CurlConverter.to_curl(method=self.method,
                                              url=self.url,
                                              params=self.params,
                                              headers=self.headers,
                                              data=self._convert_body_to_string(indent=None)
                                              if self.body is not None else None,
                                              multiline=False if Config.one_line_request else True)
        return result_string

    def _create_response(self, response: Response) -> 'RESTResponse':
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        return RESTResponse(url=response.request.url,
                            method=response.request.method,
                            path=response.request.path_url,
                            request_headers=dict(response.request.headers),
                            request_body=str(response.request.body) if response.request.body else None,
                            status_code=response.status_code,
                            reason=response.reason,
                            response_headers=dict(response.headers),
                            response_body=str(response.text),
                            elapsed_time=str(response.elapsed))

    def _log_request(self, to_allure: bool = True) -> None:
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        curl_request = self.get_curl_string()
        response_repr = self.response.get_string_repr(one_line=Config.one_line_response) if self.response is not None \
            else 'Response not available'

        self._log.debug(f'Request curl:\n{curl_request}')
        self._log.debug(f'Response string:\n{response_repr}')

        if Config.print_request_to_std_out:
            print(curl_request)
        if Config.print_response_to_std_out:
            print(response_repr)

        if to_allure:
            with allure.step(f'{self.method} {self.url}'):
                try:
                    allure.attach(curl_request, f'Request curl', allure.attachment_type.TEXT)
                    allure.attach(response_repr, f'Response', allure.attachment_type.TEXT)
                except Exception:
                    self._log.warning('Failed to attach data to allure report', stack_info=True)

    def send(self, to_allure: bool = True) -> 'RESTResponse':
        """
        Sends the request and returns the response.
        :param to_allure: flag to enable/disable allure reporting for this request
        :return: response object
        """
        self._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                        f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        try:
            response = requests.request(self.method,
                                        self.url,
                                        params=self.params,
                                        headers=self.headers,
                                        data=self._convert_body_to_string() if self.body else None,
                                        hooks=self.hooks,
                                        timeout=self.timeout)
            rest_response = self._create_response(response)
            self.response = rest_response
            return rest_response
        except Exception:
            self._log.error(f'Failed to send request "{self.method} {self.url}"', exc_info=True)
            raise Exception(f'Failed to send request "{self.method} {self.url}"')
        finally:
            self._log_request(to_allure=to_allure)



import inspect
import json
import logging
from dataclasses import is_dataclass
from typing import Any

from jto import JTOConverter

from trest.utils import is_json_string


class ObjectConverter:
    _log = logging.getLogger(__name__)

    @classmethod
    def to_string(cls, obj: Any, indent: int = None):
        """
        Converts an object to a string representation.
        :param obj: object to convert. Supported types are: dict, dataclass, str, object
        :param indent: indentation level in the string representation for json
        :return:
        """
        cls._log.debug(f'function "{inspect.currentframe().f_code.co_name}" '
                       f'called with args "{inspect.getargvalues(inspect.currentframe()).locals}"')
        if type(obj) == dict:
            string_repr = json.dumps(obj, indent=indent)
        elif is_dataclass(obj):
            string_repr = JTOConverter.to_json(obj)
            string_repr = json.dumps(string_repr, indent=indent)
        elif type(obj) == str:
            if is_json_string(obj):
                string_repr = json.dumps(json.loads(obj), indent=indent)
            else:
                string_repr = obj
        else:
            try:
                string_repr = json.dumps(obj, default=vars, indent=indent)
            except Exception:
                cls._log.error(f'Unexpected body type "{str(type(obj))}" was passed '
                               f'and cannot be converted to string')
                raise ValueError(f'Unexpected body type "{str(type(obj))}" was passed '
                                 f'and cannot be converted to string')
        return string_repr



from trest.curl_converter import CurlConverter


def test_curl_get():
    curl_str = CurlConverter.to_curl('GET', 'https://www.example.com')
    assert curl_str == 'curl -X GET https://www.example.com'


def test_curl_post_json():
    curl_str = CurlConverter.to_curl(method='POST', url='https://www.example.com',
                                     headers={'Content-Type': 'application/json'},
                                     data='{"key1": "value1", "key2": "value2"}')
    assert curl_str == 'curl -X POST -H "Content-Type: application/json" ' \
                       '-d \'{"key1": "value1", "key2": "value2"}\' ' \
                       'https://www.example.com'


def test_simple_params():
    curl_str = CurlConverter.to_curl(method='POST', url='https://www.example.com',
                                     params={'p1': 'v1', 'p2': 'v2'},
                                     headers={'Content-Type': 'application/json'},
                                     data='{"key1": "value1", "key2": "value2"}')
    assert curl_str == 'curl -X POST -H "Content-Type: application/json" ' \
                       '-d \'{"key1": "value1", "key2": "value2"}\' ' \
                       'https://www.example.com?p1=v1&p2=v2'


def test_list_param():
    curl_str = CurlConverter.to_curl(method='POST', url='https://www.example.com',
                                     params={'p1': 'v1', 'p2[]': ['111', '222']},
                                     headers={'Content-Type': 'application/json'},
                                     data='{"key1": "value1", "key2": "value2"}')
    assert curl_str == 'curl -X POST -H "Content-Type: application/json" ' \
                       '-d \'{"key1": "value1", "key2": "value2"}\' ' \
                       'https://www.example.com?p1=v1&p2[]=111&p2[]=222'
